//#case
#include <16F876.H>
//#device *=16


// Configure PIC to use: HS clock, no Watchdog Timer,
// no code protection, no Power Up Timer, etc.
#fuses HS,NOWDT,NOPROTECT,NOPUT, NOBROWNOUT, NOLVP, NOCPD, WRT

// USE BOOTLOADER
#ORG 0x1F00,0x1FFF //for the 8k 16F876/7
void loader() { }

// Tell compiler clock is 20MHz.  This is required for delay_ms()
// and for all serial I/O (such as printf(...).  These functions
// use software delay loops, so the compiler needs to know the
// processor speed.
//
#use DELAY(clock=20000000)

// Hardware definitions
#define         OE      PIN_C1
#define         DS      PIN_C0
#define         CK      PIN_C2

#define         C4      PIN_B7
#define         C3      PIN_B6
#define         C2      PIN_B3
#define         C1      PIN_B4
#define         C0      PIN_B5

// Hard parameters
#define W 10   
#define H 14

#define SIG_DELAY 4  // Anything lower doesn't seem to work
#define REFRESH_PERIOD 7167  // = ((20 MHz / 4 cyc per inst.) / (60 frames per sec)) / 5 cols per frame

// Colors
#define OFF  0
#define HALF 1
#define ON   2

//Movie data
#define FRAME_DELAY 30


//  Be sure N_LETTERS matches the number of letters in "message"
//  "message" is the scrolling message as an array of chars.
//  only lowercase a-z are defined. Feel free to add more characters.
#define N_LETTERS 45
const int message[N_LETTERS] = {'j','a','m','e','s',' ','t','i','c','h','e','n','o','r',' ',' ',
                                'e','r','i','c',' ','r','o','s','e','n','b','a','u','m',' ',' ',
                                'k','e','v','i','n',' ','c','h','e','n',' ',' ',' '};

int l1;
int l1width;
int l1start;

int l2;
int l2width;
int l2start;

int l3;
int l3width;
int l3start;

int l4;
int l4width;
int l4start;

int letterPtr;


// Buffer storage
int frameCol;
short inBufRefresh;
int  bufNum;
unsigned long buf0[10];
unsigned long buf1[10];

unsigned long offBuf0[10];
unsigned long offBuf1[10];


void clearOffScreen() {
	int i;
	for (i = 0; i < 10; i++) {
		offBuf0[i] = 0L;
		offBuf1[i] = 0L;
	}
}


void swapBuffs() {
	int i;
	
	for (i = 0; i < 10; i++) {
		buf0[i] = offBuf0[i];
		buf1[i] = offBuf1[i];
	}
}

void point(int x, int y, int color) {
//  int x, y;

  if (color == OFF) {
    offBuf0[x] &= ~(1L<<(long)y);
    offBuf1[x] &= ~(1L<<(long)y);
  }
  else if (color == HALF) {
    offBuf0[x] |= (1L<<(long)y);
    offBuf1[x] &= ~(1L<<(long)y);
  }
  else {
    offBuf0[x] |= (1L<<(long)y);
    offBuf1[x] |= (1L<<(long)y);
  }
}

long getLetterCol(int letter, int letterCol) {
	long bits[9];
	long result;
	int i;
	
	switch (letter) {
		case 'a':
			bits[0] = 0b0000000000000000L;
			bits[1] = 0b0000000000000000L;
			bits[2] = 0b0000000000011110L;
			bits[3] = 0b0000000000100000L;
			bits[4] = 0b0000000000111110L;
			bits[5] = 0b0000000000100001L;
			bits[6] = 0b0000000000111110L;
			bits[7] = 0b0000000000000000L;
			bits[8] = 0b0000000000000000L;
			break;
		case 'b':
			bits[0] = 0b0000000000000001L;
			bits[1] = 0b0000000000000001L;
			bits[2] = 0b0000000000011111L;
			bits[3] = 0b0000000000100001L;
			bits[4] = 0b0000000000100001L;
			bits[5] = 0b0000000000100001L;
			bits[6] = 0b0000000000011111L;
			bits[7] = 0b0000000000000000L;
			bits[8] = 0b0000000000000000L;
			break;
		case 'c':
			bits[0] = 0b0000000000000000L;
			bits[1] = 0b0000000000000000L;
			bits[2] = 0b0000000000011110L;
			bits[3] = 0b0000000000000001L;
			bits[4] = 0b0000000000000001L;
			bits[5] = 0b0000000000000001L;
			bits[6] = 0b0000000000011110L;
			bits[7] = 0b0000000000000000L;
			bits[8] = 0b0000000000000000L;
			break;
		case 'd':
			bits[0] = 0b0000000000100000L;
			bits[1] = 0b0000000000100000L;
			bits[2] = 0b0000000000111110L;
			bits[3] = 0b0000000000100001L;
			bits[4] = 0b0000000000100001L;
			bits[5] = 0b0000000000100001L;
			bits[6] = 0b0000000000111110L;
			bits[7] = 0b0000000000000000L;
			bits[8] = 0b0000000000000000L;
			break;
		case 'e':
			bits[0] = 0b0000000000000000L;
			bits[1] = 0b0000000000000000L;
			bits[2] = 0b0000000000011110L;
			bits[3] = 0b0000000000100001L;
			bits[4] = 0b0000000000111111L;
			bits[5] = 0b0000000000000001L;
			bits[6] = 0b0000000000111110L;
			bits[7] = 0b0000000000000000L;
			bits[8] = 0b0000000000000000L;
			break;
		case 'f':
			bits[0] = 0b0000000000011100L;
			bits[1] = 0b0000000000000010L;
			bits[2] = 0b0000000000011111L;
			bits[3] = 0b0000000000000010L;
			bits[4] = 0b0000000000000010L;
			bits[5] = 0b0000000000000010L;
			bits[6] = 0b0000000000000010L;
			bits[7] = 0b0000000000000000L;
			bits[8] = 0b0000000000000000L;
			break;
		case 'g':
			bits[0] = 0b0000000000000000L;
			bits[1] = 0b0000000000000000L;
			bits[2] = 0b0000000000111110L;
			bits[3] = 0b0000000000100001L;
			bits[4] = 0b0000000000100001L;
			bits[5] = 0b0000000000100001L;
			bits[6] = 0b0000000000111110L;
			bits[7] = 0b0000000000100000L;
			bits[8] = 0b0000000000011110L;
			break;
		case 'h':
			bits[0] = 0b0000000000000001L;
			bits[1] = 0b0000000000000001L;
			bits[2] = 0b0000000000000001L;
			bits[3] = 0b0000000000011111L;
			bits[4] = 0b0000000000100001L;
			bits[5] = 0b0000000000100001L;
			bits[6] = 0b0000000000100001L;
			bits[7] = 0b0000000000000000L;
			bits[8] = 0b0000000000000000L;
			break;
		case 'i':
			bits[0] = 0b0000000000000001L;
			bits[1] = 0b0000000000000000L;
			bits[2] = 0b0000000000000001L;
			bits[3] = 0b0000000000000001L;
			bits[4] = 0b0000000000000001L;
			bits[5] = 0b0000000000000001L;
			bits[6] = 0b0000000000000001L;
			bits[7] = 0b0000000000000000L;
			bits[8] = 0b0000000000000000L;
			break;
		case 'j':
			bits[0] = 0b0000000000001000L;
			bits[1] = 0b0000000000001000L;
			bits[2] = 0b0000000000001000L;
			bits[3] = 0b0000000000001000L;
			bits[4] = 0b0000000000001000L;
			bits[5] = 0b0000000000001000L;
			bits[6] = 0b0000000000000111L;
			bits[7] = 0b0000000000000000L;
			bits[8] = 0b0000000000000000L;
			break;
		case 'k':
			bits[0] = 0b0000000000000001L;
			bits[1] = 0b0000000000000001L;
			bits[2] = 0b0000000000110001L;
			bits[3] = 0b0000000000001001L;
			bits[4] = 0b0000000000000111L;
			bits[5] = 0b0000000000001001L;
			bits[6] = 0b0000000000110001L;
			bits[7] = 0b0000000000000000L;
			bits[8] = 0b0000000000000000L;
			break;
		case 'l':
			bits[0] = 0b0000000000000001L;
			bits[1] = 0b0000000000000001L;
			bits[2] = 0b0000000000000001L;
			bits[3] = 0b0000000000000001L;
			bits[4] = 0b0000000000000001L;
			bits[5] = 0b0000000000000001L;
			bits[6] = 0b0000000000000001L;
			bits[7] = 0b0000000000000000L;
			bits[8] = 0b0000000000000000L;
			break;
		case 'm':
			bits[0] = 0b0000000000000000L;
			bits[1] = 0b0000000000000000L;
			bits[2] = 0b0000001111011111L;
			bits[3] = 0b0000010000100001L;
			bits[4] = 0b0000010000100001L;
			bits[5] = 0b0000010000100001L;
			bits[6] = 0b0000010000100001L;
			bits[7] = 0b0000000000000000L;
			bits[8] = 0b0000000000000000L;
			break;
		case 'n':
			bits[0] = 0b0000000000000000L;
			bits[1] = 0b0000000000000000L;
			bits[2] = 0b0000000000011111L;
			bits[3] = 0b0000000000100001L;
			bits[4] = 0b0000000000100001L;
			bits[5] = 0b0000000000100001L;
			bits[6] = 0b0000000000100001L;
			bits[7] = 0b0000000000000000L;
			bits[8] = 0b0000000000000000L;
			break;
		case 'o':
			bits[0] = 0b0000000000000000L;
			bits[1] = 0b0000000000000000L;
			bits[2] = 0b0000000000011110L;
			bits[3] = 0b0000000000100001L;
			bits[4] = 0b0000000000100001L;
			bits[5] = 0b0000000000100001L;
			bits[6] = 0b0000000000011110L;
			bits[7] = 0b0000000000000000L;
			bits[8] = 0b0000000000000000L;
			break;
		case 'p':
			bits[0] = 0b0000000000000000L;
			bits[1] = 0b0000000000000000L;
			bits[2] = 0b0000000000011111L;
			bits[3] = 0b0000000000100001L;
			bits[4] = 0b0000000000100001L;
			bits[5] = 0b0000000000100001L;
			bits[6] = 0b0000000000011111L;
			bits[7] = 0b0000000000000001L;
			bits[8] = 0b0000000000000001L;
			break;
		case 'q':
			bits[0] = 0b0000000000000000L;
			bits[1] = 0b0000000000000000L;
			bits[2] = 0b0000000000111110L;
			bits[3] = 0b0000000000100001L;
			bits[4] = 0b0000000000100001L;
			bits[5] = 0b0000000000100001L;
			bits[6] = 0b0000000000111110L;
			bits[7] = 0b0000000000100000L;
			bits[8] = 0b0000000000100000L;
			break;
		case 'r':
			bits[0] = 0b0000000000000000L;
			bits[1] = 0b0000000000000000L;
			bits[2] = 0b0000000000011101L;
			bits[3] = 0b0000000000000011L;
			bits[4] = 0b0000000000000001L;
			bits[5] = 0b0000000000000001L;
			bits[6] = 0b0000000000000001L;
			bits[7] = 0b0000000000000000L;
			bits[8] = 0b0000000000000000L;
			break;
		case 's':
			bits[0] = 0b0000000000000000L;
			bits[1] = 0b0000000000000000L;
			bits[2] = 0b0000000000011110L;
			bits[3] = 0b0000000000000001L;
			bits[4] = 0b0000000000001110L;
			bits[5] = 0b0000000000010000L;
			bits[6] = 0b0000000000001111L;
			bits[7] = 0b0000000000000000L;
			bits[8] = 0b0000000000000000L;
			break;
		case 't':
			bits[0] = 0b0000000000000010L;
			bits[1] = 0b0000000000000010L;
			bits[2] = 0b0000000000011111L;
			bits[3] = 0b0000000000000010L;
			bits[4] = 0b0000000000000010L;
			bits[5] = 0b0000000000000010L;
			bits[6] = 0b0000000000011100L;
			bits[7] = 0b0000000000000000L;
			bits[8] = 0b0000000000000000L;
			break;
		case 'u':
			bits[0] = 0b0000000000000000L;
			bits[1] = 0b0000000000000000L;
			bits[2] = 0b0000000000100001L;
			bits[3] = 0b0000000000100001L;
			bits[4] = 0b0000000000100001L;
			bits[5] = 0b0000000000100001L;
			bits[6] = 0b0000000000111110L;
			bits[7] = 0b0000000000000000L;
			bits[8] = 0b0000000000000000L;
			break;
		case 'v':
			bits[0] = 0b0000000000000000L;
			bits[1] = 0b0000000000000000L;
			bits[2] = 0b0000000001000001L;
			bits[3] = 0b0000000001000001L;
			bits[4] = 0b0000000000100010L;
			bits[5] = 0b0000000000010100L;
			bits[6] = 0b0000000000001000L;
			bits[7] = 0b0000000000000000L;
			bits[8] = 0b0000000000000000L;
			break;
		case 'w':
			bits[0] = 0b0000000000000000L;
			bits[1] = 0b0000000000000000L;
			bits[2] = 0b0000000100010001L;
			bits[3] = 0b0000000100010001L;
			bits[4] = 0b0000000100010001L;
			bits[5] = 0b0000000010101010L;
			bits[6] = 0b0000000001000100L;
			bits[7] = 0b0000000000000000L;
			bits[8] = 0b0000000000000000L;
			break;
		case 'x':
			bits[0] = 0b0000000000000000L;
			bits[1] = 0b0000000000000000L;
			bits[2] = 0b0000000000100001L;
			bits[3] = 0b0000000000010010L;
			bits[4] = 0b0000000000001100L;
			bits[5] = 0b0000000000010010L;
			bits[6] = 0b0000000000100001L;
			bits[7] = 0b0000000000000000L;
			bits[8] = 0b0000000000000000L;
			break;
		case 'y':
			bits[0] = 0b0000000000000000L;
			bits[1] = 0b0000000000000000L;
			bits[2] = 0b0000000000100001L;
			bits[3] = 0b0000000000100001L;
			bits[4] = 0b0000000000100001L;
			bits[5] = 0b0000000000100001L;
			bits[6] = 0b0000000000111110L;
			bits[7] = 0b0000000000100000L;
			bits[8] = 0b0000000000011110L;
			break;
		case 'z':
			bits[0] = 0b0000000000000000L;
			bits[1] = 0b0000000000000000L;
			bits[2] = 0b0000000000011111L;
			bits[3] = 0b0000000000001000L;
			bits[4] = 0b0000000000000100L;
			bits[5] = 0b0000000000000010L;
			bits[6] = 0b0000000000011111L;
			bits[7] = 0b0000000000000000L;
			bits[8] = 0b0000000000000000L;
			break;
		default:
			bits[0] = 0b0000000000000000L;
			bits[1] = 0b0000000000000000L;
			bits[2] = 0b0000000000000000L;
			bits[3] = 0b0000000000000000L;
			bits[4] = 0b0000000000000000L;
			bits[5] = 0b0000000000000000L;
			bits[6] = 0b0000000000000000L;
			bits[7] = 0b0000000000000000L;
			bits[8] = 0b0000000000000000L;
			break;
	}
	
	result = 0;
	for (i = 0; i < 9; i++) {
		if (bit_test(bits[i], letterCol)) {
			bit_set(result, i);
		}
	}
	return result;
}

int getWidth(int letter) {
	
	switch (letter) {
		case 'a':
			return 6;
		case 'b':
			return 6;
		case 'c':
			return 5;
		case 'd':
			return 6;
		case 'e':
			return 6;
		case 'f':
			return 5;
		case 'g':
			return 6;
		case 'h':
			return 6;
		case 'i':
			return 1;
		case 'j':
			return 4;
		case 'k':
			return 6;
		case 'l':
			return 1;
		case 'm':
			return 11;
		case 'n':
			return 6;
		case 'o':
			return 6;
		case 'p':
			return 6;
		case 'q':
			return 6;
		case 'r':
			return 5;
		case 's':
			return 5;
		case 't':
			return 5;
		case 'u':
			return 6;
		case 'v':
			return 7;
			break;
		case 'w':
			return 9;
		case 'x':
			return 6;
		case 'y':
			return 6;
		case 'z':
			return 5;
		default:
			return 4;
	}
}

void putLetterColInBufferCol(int letter, int letterCol, int bufferCol) {
	long letterBits;
	int i;
	
	if (bufferCol >= H)
		return;

	letterBits = getLetterCol(letter, letterCol);
	for (i = 0; i < 9; i++) {
		if (bit_test(letterBits, i)) {
			point((W - i)-1, bufferCol, ON);
		}
	}
}

void advanceLetter() {
	letterPtr = (letterPtr + 1) % N_LETTERS;

	l1start = l2start;
	l1width = l2width;
	l1 = l2;
	
	l2start = l3start;
	l2width = l3width;
	l2 = l3;
	
	l3start = l4start;
	l3 = l4;
	l3width = l4width;
	
	l4start = l3start + l3width + 2;
	l4 = message[letterPtr];
	l4width = getWidth(l4);
}

int drawFourLetters() {
	int lx;
	short int usedL1;
	
	usedL1 = FALSE;
	for(lx = 0; lx < l1width; lx++) {
		if (l1start + lx < 100) { // same as > 0
			usedL1 = TRUE;
			putLetterColInBufferCol(l1, lx, lx + l1start);
		}
	}
	
	if (! usedL1) {
		advanceLetter();
		return FALSE;
	}
	
	for(lx = 0; lx < l2width; lx++) {
		putLetterColInBufferCol(l2, lx, lx + l2start);
	}

	for(lx = 0; lx < l3width; lx++) {
		putLetterColInBufferCol(l3, lx, lx + l3start);
	}


	for(lx = 0; lx < l4width; lx++) {
		putLetterColInBufferCol(l4, lx, lx + l4start);
	}

	return TRUE;
}


void initMovie() {
	clearOffScreen();
	swapBuffs();
	letterPtr = 0;
	
	l1 = message[0];
	l1width = getWidth(l1);
	l1start = 0;
	
	l2 = message[1];
	l2width = getWidth(l2);
	l2start = l1width + 2;
	
	l3 = message[2];
	l3width = getWidth(l3);
	l3start = l2start + l2width + 2;
	
	l4 = message[3];
	l4width = getWidth(l4);
	l4start = l3start + l3width + 2;
	
	letterPtr = 4 % N_LETTERS;
}



void nextFrame() {
	clearOffScreen();
	if (!drawFourLetters())
		drawFourLetters();
	
	swapBuffs();

	l1start--;
	l2start--;
	l3start--;
	l4start--;
}

void runMovie() {
	while (1) {
		nextFrame();
		delay_ms(FRAME_DELAY);
	}
}


void clockSR() {
  output_low(CK);
  delay_us(SIG_DELAY);
  output_high(CK);
  delay_us(SIG_DELAY);
}


void setupCols() {
  if (frameCol == 0) {
    output_high(C0);
  }
  else {
    output_low(C0);
  }
 
  if (frameCol == 1) {
    output_high(C1);
  }
  else {
    output_low(C1);
  }
 
  if (frameCol == 2) {
    output_high(C2);
  } 
  else {
    output_low(C2);
  }  
 
  if (frameCol == 3) {
    output_high(C3);
  }
  else {
    output_low(C3);
  }
 
  if (frameCol == 4) {
    output_high(C4);
  }
  else {
    output_low(C4);
  }

  output_low(OE); // oe
}


void drawFrameCol(unsigned long* buf) {
    int i;
    long mask;
	  
    output_high(OE);        // oe
    delay_us(SIG_DELAY);

  // TOP LEFT
  mask = 0b0000000010000000;
  for (i = 0; i < 7; i++) {

    if ((buf[frameCol] & mask) != 0)	// Light this row
		output_low(DS);
	else 
		output_high(DS);
	delay_us(SIG_DELAY);

	clockSR();

	mask <<= 1;
  }
  clockSR();


  // BOTTOM LEFT
  mask = 0b0000000010000000;
  for (i = 0; i < 7; i++) {

	if ((buf[frameCol + 5] & mask) != 0)	// Light this row
		output_low(DS);
	else 
		output_high(DS);
	delay_us(SIG_DELAY);

	clockSR();

	mask <<= 1;
  }
  clockSR();

  // BOTTOM RIGHT
  mask = 0b0000000000000001;
  for (i = 0; i < 7; i++) {
    if ((buf[frameCol + 5] & mask) != 0)	// Light this row
	  output_low(DS);
	else 
	  output_high(DS);
	delay_us(SIG_DELAY);

	clockSR();

	mask <<= 1;
  }
  clockSR();


  // TOP RIGHT
  mask = 0b0000000000000001;
  for (i = 0; i < 7; i++) {

    if ((buf[frameCol] & mask) != 0)	// Light this row
	  output_low(DS);
	else 
	  output_high(DS);
    delay_us(SIG_DELAY);

	clockSR();

	mask <<= 1;
  }
  clockSR();

      
  clockSR();
  setupCols();
}


void initTimer() {
  frameCol = 0;
  CCP_1 = REFRESH_PERIOD;
  enable_interrupts(INT_CCP1);
  enable_interrupts(GLOBAL);
  setup_ccp1(CCP_COMPARE_RESET_TIMER);
  setup_timer_1(T1_INTERNAL);
}


void main() {
    clearOffScreen();
	  initMovie();
		initTimer();
		runMovie();
}

#INT_CCP1
void DRAW_COL() {
  if (frameCol == 4) {
    if (bufNum == 3)
      bufNum = 0;
    else
      bufNum++;
    frameCol = 0;
  }
  else
    frameCol++;

  if (bufNum == 0)
    drawFrameCol(buf0);
  else 
    drawFrameCol(buf1);
  
}

